"""
Dataclass representing one TRC20 token transfer.
"""
from dataclasses import asdict, dataclass, field, fields
from typing import Any, Dict, List, Optional, Tuple, Union

import pendulum
from rich.text import Text

from trongrid_extractoor.config import log
from trongrid_extractoor.helpers.address_helpers import coerce_to_base58
from trongrid_extractoor.helpers.rich_helpers import color_picker
from trongrid_extractoor.helpers.string_constants import ADDRESS, DATA, EVENT_NAME, RESULT, RESULT_TYPE, UINT256

KEYS_TO_STRIP_FROM_FLAT_DICT = ['datetime', 'ms_from_epoch', 'raw_event']


@dataclass(kw_only=True)
class TronEvent:
    event_name: str
    token_address: str  # TODO: Should be contract_address
    transaction_id: str
    event_index: int
    ms_from_epoch: int
    raw_event: Optional[Dict[str, Union[dict, str, float, int]]] = None
    block_number: Optional[int] = None
    datetime: Optional[pendulum.DateTime] = None  # Derived field; not passed in constructor

    def __post_init__(self):
        # Type coercion
        self.block_number = int(self.block_number) if self.block_number else None
        self.ms_from_epoch = int(float(self.ms_from_epoch))
        self.event_index = int(self.event_index)

        # Computed fields
        self.seconds_from_epoch = int(self.ms_from_epoch / 1000.0)
        self.datetime = pendulum.from_timestamp(self.seconds_from_epoch, pendulum.tz.UTC)
        self.unique_id = f"{self.transaction_id}/{self.event_index}"

    @classmethod
    def from_event_dict(cls, row: Dict[str, Union[str, float, int]]) -> 'TronEvent':
        """Build an event from the json data returned by Trongrid."""
        return cls(
            event_name=row[EVENT_NAME],
            token_address=row['contract_address'],
            ms_from_epoch=row['block_timestamp'],
            block_number=row['block_number'],
            transaction_id=row['transaction_id'],
            event_index=row['event_index'],
            raw_event=row
        )

    def event_properties(self) -> Dict[str, bool|float|int|str]:
        """
        Normalize the 'result' dict which is where the event specific properties live.
        Other conversions:
           - hex addresses converted to base58
           - booleans and numbers coerced to native types
        """
        if self.raw_event is None:
            raise ValueError(f"No 'raw_event' to normalize!")

        result = self.raw_event[RESULT]
        result_types = self.raw_event[RESULT_TYPE]

        if 2 * len(result_types) != len(result):
            log.warning(f"Found {len(result_types)} result types for {len(result)} results! Naming may be incorrect.")

        return {
            arg_name: self._normalize_result_value(arg_type, result[arg_name])
            for arg_name, arg_type in result_types.items()
        }

    def to_properties_dict(self) -> Dict[str, bool|float|int|str]:
        """Convert to a flat key/value store with all relevant properties."""
        base_dict = asdict(self)
        base_dict['block_written_at'] = self.datetime
        base_dict['caller_contract_address'] = self.raw_event['caller_contract_address']
        base_dict['contract_address'] = base_dict['token_address']
        del base_dict['token_address']

        for key in KEYS_TO_STRIP_FROM_FLAT_DICT:
            del base_dict[key]

        base_dict.update(self.event_properties())
        return base_dict

    def _normalize_result_value(self, arg_type: str, value: str|int|float) -> str|int|float:
        if not isinstance(value, str):
            return value

        if arg_type == ADDRESS:
            return coerce_to_base58(value)
        elif arg_type == UINT256:
            return int(value)
        elif arg_type == 'bool':
            if value.lower() == 'true':
                return True
            elif value.lower() == 'false':
                return False
            else:
                raise ValueError(f"Cannot coerce '{value}' to boolean")
        else:
            log.warning(f"Unknown type '{arg_type}', cannot coerce '{value}' so returning string...")
            return value

    def __str__(self) -> str:
        msg = f"{self.event_name}: {self.token_address}, ID: {self.transaction_id}/{self.event_index}"
        return msg

    def __rich__(self) -> Text:
        color = color_picker.pick_color(self.event_name)
        txt = Text('[').append(str(self.datetime)).append('] ').append(self.event_name, style=color).append('\n')

        for k, v in self.event_properties().items():
            txt.append(f"    {k}", style='white').append(": ").append(str(v) + '\n', style='color(217)')

        return txt.append('\n')
